VERSION = "1.0.1"
PROJECT = "plenty"

URL = "https://github.com/zlj-zz/plenty"
URI = URL

AUTHOR = "Zachary Zhang"
EMAIL = "zlj19971222@outlook.com"

LICENSE = "MIT"
COPYRIGHT = "Copyright (c) 2022 Zachary"

DESC = "Terminal diversification class library."
